#!/usr/bin/perl

# Article:
# [Pereira and Jamhour, 2011]
# Metodo Heuristico para Rotular Grupos em Sistema de Deteccao de Intrusao baseado em Anomalia

# URL: http://ids.celepar.pr.gov.br/scripts

# Labeling

use strict;

my $example = "Example(1): perl $0 dts.p00.txt.dsn.100.cls portnoy 0.15\nResults: dts.p00.txt.dsn.100.cls.portnoy.0.15.lbl\n";
$example .= "Example(2): perl $0 dts.p00.txt.dsn.100.cls zhong 0.115\nResults: dts.p00.txt.dsn.100.cls.zhong.0.115.lbl\n";
$example .= "Example(3): perl $0 dts.p00.txt.dsn.100.cls proposed 0.05\nResults: dts.p00.txt.dsn.100.cls.proposed.0.05.lbl\n";
my $file = $ARGV[0] || die $example;
my $method = $ARGV[1] || "portnoy"; # portnoy, zhong or proposed
my $param = $ARGV[2] || 0.1; # perc (portnoy,zhong) cutline (proposed)

my $file_cls = "$file";
my $file_dsc = substr($file_cls,0,length($file_cls)-3)."dsc"; 
my $file_lbl = $file.".".$method.".".$param.".lbl";

my @cls = &load_clusters($file_cls);

my @lbl;
my @dsc;
if ($method eq "portnoy") {
   @lbl = &labeling_portnoy($param,@cls);
} elsif ($method eq "zhong") {
   @lbl = &labeling_zhong($param,@cls);
} elsif ($method eq "proposed") {
   @dsc = &load_dsclusters($file_dsc);
   my $yp = 1; # yp - popular
   my $yc = 3; # yc - reliable
   my $limitlabel = 1.0; # reputation < 1.0 is attack
   @lbl = &labeling_proposed($param,$yp,$yc,$limitlabel,\@cls,\@dsc);
} else {
   die "Error: method (portnoy, zhong or proposed)\n";
}

&save_clusters($file_lbl,@lbl);
print "done - $file_lbl\n";

##############################################################################
sub load_clusters {
   my $file = shift;
   my @cls;
   my ($k,$line);
   my $fline = 1;
   open (FILE, $file);
   while ($line = <FILE>) {
      if (! $fline) {
         chomp($line);
         @{$cls[$k]} = split(/ +/,$line);   
         $k++;
      } else {
         $fline = 0;
      }
   } 
   close (FILE);
   return @cls;
}
##############################################################################

 
##############################################################################
sub load_dsclusters {
   my $file = shift;
   my @dsc;
   my ($i,$line);
   my $fline = 1;
   open (FILE, $file);
   while ($line = <FILE>) {
      if (! $fline) {
         chomp($line);
         @{$dsc[$i]} = split(/ +/,$line);   
         $i++;
      } else {
         $fline = 0;
      }
   } 
   close (FILE);
   return @dsc;
}
##############################################################################


##############################################################################
# Save labeled clusters
sub save_clusters {
   my $file = shift;
   my @lbl = @_;
   my ($i,$line);
   open (FILE, "> $file");
   print FILE "id requests perc result_lbl label up:alpha num other uq:alpha num other ho:alpha num other ua:alpha num other ck:alpha num other rf:alpha num other ct:alpha num other\n";
   for ($i = 0; $i < scalar(@lbl); $i++) {
      $line = join(" ",@{$lbl[$i]});
      print FILE $line."\n"; 
   }
   close (FILE);
}
##############################################################################


##############################################################################
sub labeling_portnoy {
   my $perc = shift;
   my @cls = @_;
   my %sort;
   my @sort;
   my @lbl;
   my ($k,$n,$sum,$tot_perc);
   $sum = 0;
   for ($k = 0; $k < scalar(@cls); $k++) {
      $sum += $cls[$k][1];
   }
   for ($k = 0; $k < scalar(@cls); $k++) {
      $sort{$k} = $cls[$k][1] / $sum;
   } 
   $n = 0;
   foreach $k (sort {$sort{$b} <=> $sort{$a}} keys %sort) {
      $sort[$n] = $k;
      $n++;
   }
   $tot_perc = 0;
   for ($n = 0; $n < scalar(@sort); $n++) {
      $k = $sort[$n];
      $tot_perc = ($n + 1) / scalar(@cls);
      push(@{$lbl[$n]},@{$cls[$k]}[0..1]);
      push(@{$lbl[$n]},sprintf("%3.6f",$sort{$k})); 
      push(@{$lbl[$n]},sprintf("%3.6f",$tot_perc)); 
      if ($tot_perc <= $perc) {
         push(@{$lbl[$n]},"NORMAL"); 
      } else {
         push(@{$lbl[$n]},"ATTACK");
      }
      push(@{$lbl[$n]},@{$cls[$k]}[3..23]);
   }
   return @lbl;
}
##############################################################################


##############################################################################
sub labeling_zhong {
   my $perc = shift;
   my @cls = @_;
   my %sort;
   my @sort;
   my @lbl;
   my ($k,$j,$l,$n,$max,$max_k,$k_perc,$sum,$sum_perc,$distance,$euc_sum);
   $sum = 0;
   $max = 0;
   $max_k = 0;  # largest cluster
   for ($k = 0; $k < scalar(@cls); $k++) {
      if ($cls[$k][1] > $max) {
         $max = $cls[$k][1];
         $max_k = $k;
      }
      $sum += $cls[$k][1]; 
   }
   for ($k = 0; $k < scalar(@cls); $k++) {
      $distance = 0;
      for ($l = 3; $l <= 21; $l += 3) {
         $euc_sum = 0;
         for ($j = $l; $j < $l + 3; $j++) {
            $euc_sum += (($cls[$k][$j] - $cls[$max_k][$j]) ** 2);
         }
         $distance += sqrt($euc_sum);
      }
      $sort{$k} = $distance;
   }
   $n = 0;
   foreach $k (sort {$sort{$b} <=> $sort{$a}} keys %sort) {
      $sort[$n] = $k;
      $n++;
   }
   $sum_perc = 0;
   for ($n = 0; $n < scalar(@sort); $n++) {
      $k = $sort[$n];
      $k_perc = ($cls[$k][1] / $sum);
      $sum_perc += $k_perc;
      push(@{$lbl[$n]},@{$cls[$k]}[0..1]);
      push(@{$lbl[$n]},sprintf("%3.6f",$k_perc)); 
      push(@{$lbl[$n]},$sort{$k}); 
      if ($sum_perc > $perc) {
         push(@{$lbl[$n]},"NORMAL"); 
      } else {
         push(@{$lbl[$n]},"ATTACK");
      }
      push(@{$lbl[$n]},@{$cls[$k]}[3..23]);
   }
   return @lbl;
}
##############################################################################


##############################################################################
sub labeling_proposed {
   my ($cutline,$yp,$yc,$limitlabel,$ref_cls,$ref_dsc) = @_;
   my @cls = @{$ref_cls};
   my @dsc = @{$ref_dsc};
   my $qtclusters = scalar(@cls);
   my @dsc2;
   my @indexes;
   my %ipspopular;
   my %ipsreliable;
   my $qtclusterspopular = 0; 
   # Popular clusters
   for (my $k = $qtclusters - 1; $k >= 0; $k--) {
      print "popular k $k\n";
      my %ipscluster;
      my $totreqcluster = 0;
      for (my $i = scalar(@dsc) - 1; $i >= 0; $i--) {
         if ($dsc[$i][0] == $k) {
            $ipscluster{$dsc[$i][1]}++;
            $totreqcluster++;
            push(@dsc2,splice(@dsc,$i,1));
         }
      }
      my $ipsclusterzero = 1;
      my $qtipszero = 0;
      my $subtracttotreqcluster = 0;
      while ($ipsclusterzero) {
         $ipsclusterzero = 0;
         $subtracttotreqcluster = 0;
         foreach my $ipcluster (keys (%ipscluster)) {
            my $percip = ($totreqcluster > 0) ? $ipscluster{$ipcluster} / $totreqcluster : 0;
            if ($percip > $cutline) {
               $subtracttotreqcluster += $ipscluster{$ipcluster};
               $ipscluster{$ipcluster} = 0;
               $ipsclusterzero = 1;
               $qtipszero++;
            }
         }
         $totreqcluster -= $subtracttotreqcluster;
      }
      my $qtipscluster = scalar(keys (%ipscluster));
      my $popularindex = ($qtipscluster > 0) ? ($qtipscluster - $qtipszero) / $qtipscluster : 0;
      $indexes[$k][0] = $popularindex;
      if ($popularindex > 0) {
         $qtclusterspopular++;
         foreach my $ipcluster (keys (%ipscluster)) {
            if ($ipscluster{$ipcluster} > 0) {
               $ipspopular{$ipcluster} += $popularindex;
            }
         }
      }
   }
   # Find popular IP
   my $punctmostpopularip = 0;
   foreach my $ippopular (keys (%ipspopular)) {
      my $punctpopular = $ipspopular{$ippopular} / $qtclusterspopular;
      if ($punctpopular > $punctmostpopularip) {
         $punctmostpopularip = $punctpopular;
      }
   }
   # IP reliability
   foreach my $ippopular (keys (%ipspopular)) {
      $ipsreliable{$ippopular} = ($ipspopular{$ippopular} / $qtclusterspopular) / $punctmostpopularip;
   }
   # Clusters reliability
   for (my $k = $qtclusters - 1; $k >= 0; $k--) {
      print "reliable k $k\n";
      my %ipscluster;
      for (my $i = scalar(@dsc2) -1; $i >= 0; $i--) {
         if ($dsc2[$i][0] == $k) {
            $ipscluster{$dsc2[$i][1]}++;
            splice(@dsc2,$i,1);
         }
      }
      my $sumreliablecluster = 0;
      my $qtipscluster = 0;
      foreach my $ipcluster (keys (%ipscluster)) {
         $qtipscluster++;
         if ($ipsreliable{$ipcluster}) {
            $sumreliablecluster += $ipsreliable{$ipcluster};
         }
      }
      my $reliableindex = ($qtipscluster > 0) ? $sumreliablecluster / $qtipscluster : 0;
      $indexes[$k][1] = $reliableindex;
   }
   # Clusters reputation and labeling
   my @lbl;
   my $sum = 0;
   for (my $k = 0; $k < scalar(@cls); $k++) {
      $sum += $cls[$k][1]; 
   }
   for (my $k = 0; $k < scalar(@cls); $k++) {
      my $k_perc = $cls[$k][1] / $sum;
      my $reputationindex = ($yp * $indexes[$k][0]) + ($yc * $indexes[$k][1]);
      push(@{$lbl[$k]},@{$cls[$k]}[0..1]);
      push(@{$lbl[$k]},sprintf("%3.6f",$k_perc)); 
      push(@{$lbl[$k]},$reputationindex); 
      if ($reputationindex >= $limitlabel) {
         push(@{$lbl[$k]},"NORMAL"); 
      } else {
         push(@{$lbl[$k]},"ATTACK");
      }
      push(@{$lbl[$k]},@{$cls[$k]}[3..23]);
   }
   return @lbl;
}
##############################################################################



